package djmil.cordacheckers.contracts;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameProposalContract implements net.corda.v5.ledger.utxo.Contract {

    private final static Logger log = LoggerFactory.getLogger(GameProposalContract.class);

    @Override
    public void verify(UtxoLedgerTransaction trx) {
        log.info("GameProposalContract.verify() called");

        requireThat(trx.getCommands().size() == 1, REQUIRE_SINGLE_COMMAND);

        switch ((trx.getCommands(GameProposalCommand.class).get(0))) {
            case CREATE: {
                requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
                requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

                GameProposalState outputState = trx.getOutputStates(GameProposalState.class).get(0);
                requireThat(outputState != null, CREATE_OUTPUT_STATE);

                requireThat(outputState.getAcquierColor() != null, NON_NULL_RECIPIENT_COLOR);
                break; }

            case ACCEPT:
                requireThat(trx.getInputContractStates().size() == 1, ACCEPT_INPUT_STATE);
                requireThat(trx.getOutputContractStates().size() == 1, ACCEPT_OUTPUT_STATE);

                GameProposalState inGameProposal = trx.getInputStates(GameProposalState.class).get(0);
                requireThat(inGameProposal != null, ACCEPT_INPUT_STATE);

                GameBoardState outGameBoard = trx.getOutputStates(GameBoardState.class).get(0);
                requireThat(outGameBoard != null, ACCEPT_INPUT_STATE);

                requireThat(outGameBoard.getParticipants().containsAll(inGameProposal.getParticipants()), ACCEPT_PARTICIPANTS);
                break; 

            case REJECT:
                requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
                requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);
                requireThat(trx.getInputStates(GameProposalState.class).get(0) != null, REJECT_INPUT_STATE);
                break;

            case CANCEL:
                requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
                requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
                requireThat(trx.getInputStates(GameProposalState.class).get(0) != null, CANCEL_INPUT_STATE);
                break;

            default:
                throw new CordaRuntimeException(UNKNOWN_COMMAND);
        }
    }
    
    private void requireThat(boolean asserted, String errorMessage) {
        if(!asserted) {
            throw new CordaRuntimeException("Failed requirement: " + errorMessage);
        }
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String UNKNOWN_COMMAND = "Unsupported command";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";
    static final String NON_NULL_RECIPIENT_COLOR = "GameProposal.recipientColor field can not be null";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";

    static final String ACCEPT_INPUT_STATE  = "Accept command should have exactly one GameProposal input state";
    static final String ACCEPT_OUTPUT_STATE = "Accept command should have exactly one GameBoard output state";
    static final String ACCEPT_PARTICIPANTS = "Accept command: GameBoard participants should math GameProposal participants";
}
