package djmil.cordacheckers.gameboard;

import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.contracts.GameBoardCommand;
import djmil.cordacheckers.states.GameBoardState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

import java.time.Duration;
import java.time.Instant;

public class MoveFlow implements ClientStartableFlow {

    private final static Logger log = LoggerFactory.getLogger(MoveFlow.class);

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService ledgerService;

    @CordaInject
    public FlowEngine flowEngine;

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        try {
            final MoveFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, MoveFlowArgs.class);
            final GameBoardCommand command = new GameBoardCommand(args.getMove());

            final StateAndRef<GameBoardState> utxoGameBoard = findUnconsumedGameBoardState(args.getGameBoardUuid());

            final GameBoardState newGameBoard = null; // TODO

            final UtxoSignedTransaction trx = prepareSignedTransaction(command, utxoGameBoard, newGameBoard);

            final SecureHash trxId = this.flowEngine
                .subFlow( new CommitSubFlow(trx) );

            return new FlowResult(newGameBoard, trxId)  // TODO return players perspective GB
                .toJsonEncodedString(jsonMarshallingService);
        } 
        catch (Exception e) {
            log.warn("GameBoardAction flow failed to process utxo request body " + requestBody + 
                        " because: " + e.getMessage());
            return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    private StateAndRef<GameBoardState> findUnconsumedGameBoardState (UUID gameProposalUuid) {
        /*
         * Get list of all unconsumed aka 'active' GameBoardState, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        return this.ledgerService
            .findUnconsumedStatesByType(GameBoardState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(gameProposalUuid))
            .reduce((a, b) -> {throw new IllegalStateException("Multiple states: " +a +", " +b);})
            .get();
    }

    @Suspendable
    private UtxoSignedTransaction prepareSignedTransaction(
        GameBoardCommand command,
        StateAndRef<GameBoardState> utxoGameProposal,
        GameBoardState newGameBoard
    ) {
        UtxoTransactionBuilder trxBuilder = ledgerService.createTransactionBuilder()
            .setNotary(utxoGameProposal.getState().getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(utxoGameProposal.getRef())
            .addOutputState(newGameBoard)
            .addCommand(command)
            .addSignatories(utxoGameProposal.getState().getContractState().getParticipants());

        return trxBuilder.toSignedTransaction();
    }

}
